// display.c
/*----------------------------------------------------------------------*/
/*! \file 
--------------------------------------------------------------------------

  	<h3> DISPLAY.C </h3>
  	Version	1.0<br>
  	Revision	$Revision: 1.5 $<br>
 	Copyright 	Thomas Wedemeyer, Email: avr@thomas-wedemeyer.de<br> 
  	Datum		$Date: 2011/08/14 21:41:24 $<br>

	<h3>Funktionen zur Anzeige auf der LED-Matrix</h3>
  

/*----------------------------------------------------------------------*/

#include <stdlib.h>
#include <stdio.h>
#include <inttypes.h>
#include <avr/io.h>
#include <avr/interrupt.h>
//#include <avr/signal.h>
#include "define.h"
#include "display.h"


/*----------------------------------------------------------------------*/
// Einbinden der externen Variablen
/*----------------------------------------------------------------------*/

extern const uint8_t LED_ROW[];	//!< Konstanten fr Zuordnung der LEDs zu Portpin
extern const uint16_t mask16[];	//!< Maske 16-Bit fr Bitoperationen 

// Display-Speicher
extern uint8_t	row[9];				//!< Zeilenarray -> Col sind Bitcodiert 
							//!  x | x | x | CL4 | CL3 | CL2 | CL1| CL0
extern struct bitfield_format	sec_o;		// Sekunden Einstelle
extern struct bitfield_format 	sec_t;		// Sekunden Zehnerstelle
extern struct bitfield_format	minute_o;	// Minuten Einstelle
extern struct bitfield_format 	minute_t;	// Minuten Zehnerstelle
extern struct bitfield_format	hour_o;		// Stunden Einstelle
extern struct bitfield_format 	hour_t;		// Stunden Zehnerstelle


/*----------------------------------------------------------------------*/
//! Ummappen vom einfachen Bitfield auf das Row-Array
//! \brief Das Ergebnis der Berechung wird direkt in den globalen ROW-Speicher 
//! geschrieben
//!	@param	bitfield	Bitmuster der aktuellen Ziffer
//!	@param	base_row	Untereste Reihe der Ziffer in der Anzeige
//! @param	type		Typ der Ziffer 9'er / 6'er
/*----------------------------------------------------------------------*/
void bitfield2rows(uint16_t bitfield,
					uint8_t	base_row,
					uint8_t	type)
{
	uint8_t	i;	// Zhler
	uint8_t init_ledmask = 0;

	if(type == 9)			// Feld mit 9 LEDS
		init_ledmask = 0x10;
	else
		init_ledmask = 0x02;	// Feld mit 6 LEDs

	// Ummappen auf ROWs
	for(i=0; i < 9; i++)
	{
		if(bitfield & mask16[i])
		{
			row[base_row - (i % 3)] |= init_ledmask >> (i /3);
		}
		else
		{
			row[base_row - (i % 3)] &= ~(init_ledmask >> (i /3));
		}
	}
}

#ifdef RANDOM_DISPLAY
/*----------------------------------------------------------------------*/
//! Setzen von LEDs im Ziffern-Bitfeld
//! \brief	Diese Funktion setzt die der Zeit entsprechende Anzahl 
//! Bits/LED in dem bergebenem Bitfeld. Welches Bit/Led gesetzt wird ist
//! von dem Inhalt in Timer2 abhngig.
//! @param	time		Zahlenwert des Ziffernfeldes
//! @param	fielddata	Struktur mit den Bitdaten zu diesem Ziffernfeld
/*----------------------------------------------------------------------*/
void set_bitfield(uint8_t time,	struct bitfield_format *fielddata)
{
	uint8_t				diff;
	uint8_t 			i;
	uint8_t				bit=0; 

	//cli();	// IRQ sperren!

	if(time < fielddata->bit_cnt)	// Wenn LEDs gelscht werden sollen...
	{
		fielddata->bit_cnt = 0;
		fielddata->reg  = 0;
	}		


	if(time == 0)						// lsche Zahlenspeicher 
	{
		fielddata->bit_cnt = 0;
		fielddata->reg  = 0;
	}
	else
	{
		diff = time - fielddata->bit_cnt;		// Anzahlbits, die gesetzt werden mssen
		fielddata->bit_cnt = time;

		if(diff == 0) return;		// Hier muss nichts gesetzt werden!


		for( i = 0; i < diff;i++)
		{
			// Zufallsgenerator ber Timer2:
			if (fielddata->type == 9)
			{ 
				bit = TCNT2 % 9;
			}
			else if (fielddata->type == 6)
			{
				bit = TCNT2 % 6;
			}
						
			//bit = 0;

			while(1)
			{
				// Umbruch bei Suche nach freiem Pltzchen fr das Bit
				if (((fielddata->type == 9) && (bit > 8)) |
					((fielddata->type == 6) && (bit > 5)))
				{
					bit = 0;
				}
				
				if(fielddata->reg & mask16[bit])
					bit++;
				else
				{
					fielddata->reg |= mask16[bit];
					break;
				}

				// Abbruch, wenn Feld voll!
				if(((fielddata->type == 6) && (fielddata->reg == 0x003F)) |
					((fielddata->type == 9) && (fielddata->reg == 0x01FF)))	
					break;
			};
		}
	}
	

	//sei(); // Freigabe IRQ
	
	bitfield2rows(fielddata->reg, fielddata->row_base, fielddata->type);
									
}

#else	

/*----------------------------------------------------------------------*/
//! Setzen von LEDs im Ziffern-Bitfeld
//! \brief	Diese Funktion setzt die der Zeit entsprechende Anzahl 
//! Bits/LED in dem bergebenem Bitfeld. Welches Bit/Led-Muster erzeigt wird 
//! ist vom Array led_mask abhngig
//! @param	time		Zahlenwert des Ziffernfeldes
//! @param	fielddata	Struktur mit den Bitdaten zu diesem Ziffernfeld
/*----------------------------------------------------------------------*/
void set_bitfield(uint8_t time,	struct bitfield_format *fielddata)
{
	const uint16_t		led_mask[10] = { 0x0000,		// 0		Aufbau LED-Feld / Bit-Nr:		
														//								9	6	3
										 0x0001,		// 1							8	5	2
										 0x0003,		// 2							7	4	1
										 0x0007,		// 3
										 0x000F,		// 4
										 0x001F,		// 5
										 0x003F,		// 6
										 0x007F,		// 7
										 0x00FF,		// 8
										 0x01FF};		// 9


	//cli();	// IRQ sperren!

	if (time == fielddata->bit_cnt)
		return;

	fielddata->bit_cnt = time;
	fielddata->reg  = led_mask[time];

	
	//sei(); // Freigabe IRQ
	
	bitfield2rows(fielddata->reg, fielddata->row_base, fielddata->type);
									
}
#endif		

/*----------------------------------------------------------------------*/
//! Initialisierung der Displayausgabe
/*----------------------------------------------------------------------*/
void init_bitfields()
{
	sec_o.reg = 0;		// Sekunden Einstelle
	sec_o.bit_cnt = 0;
	sec_o.type = 9;
	sec_o.row_base = 8;

	sec_t.reg = 0;		// Sekunden Einstelle
	sec_t.bit_cnt = 0;
	sec_t.type = 6;
	sec_t.row_base = 8;

	minute_o.reg = 0;		// Sekunden Einstelle
	minute_o.bit_cnt = 0;
	minute_o.type = 9;
	minute_o.row_base = 5;

	minute_t.reg = 0;		// Sekunden Einstelle
	minute_t.bit_cnt = 0;
	minute_t.type = 6;
	minute_t.row_base = 5;

	hour_o.reg = 0;		// Sekunden Einstelle
	hour_o.bit_cnt = 0;
	hour_o.type = 9;
	hour_o.row_base = 2;

	hour_t.reg = 0;		// Sekunden Einstelle
	hour_t.bit_cnt = 0;
	hour_t.type = 6;
	hour_t.row_base = 2;
}



/*----------------------------------------------------------------------*/
//! Setzen der LEDs einer Spalte
//! \brief Diese Funktion wird nicht mehr benutzt. Diese Ausgabe erfolgt
//! direkt in der IRQ-Routine von Timer 2
//! @param col	Aktuelle Spalte
/*----------------------------------------------------------------------*/
void set_col(uint8_t col)
{
	const uint8_t mask[]= {0x01, 0x02, 0x04, 0x08, 0x10,0x20, 0x40, 0x80}; 
	uint8_t i;
	uint8_t	port_tmp =0;

	for(i = 0; i < ROW_MAX; i++)
	{
		switch(i)	// Laden des Portregisters:
		{
			case 0:	// ROW 0
				port_tmp = PORTC;
				break;

			case 1: // Row 1 bis 5
				port_tmp = PORTD;
				break;

			case 6: // ROW 6 bis 8
				port_tmp = PORTB;
				break;
		}

		if(mask[col] & row[i])
			port_tmp &= ~LED_ROW[i];  // Ansteuerung invertiert durch PNP-Transistor
		else
			port_tmp |= LED_ROW[i];

		switch(i) // speichern der Portregisters
		{
			case 0:	// ROW 0
				PORTC = port_tmp;
				break;
			case 5: // ROW 5
				PORTD = port_tmp;
				break;
			case 8: // ROW 8
				PORTB = port_tmp;
				break;
		}
	}
}

/*----------------------------------------------------------------------*/
//! Ausgabe-Funktion fr die LED-Matrix
//! \brief Diese Funktion wird nicht mehr benutzt. Die Aufgabe wird jetzt
//! direkt in der IRQ-Routine von Timer 2 ausgefhrt.
/*----------------------------------------------------------------------*/
void refresh_output()
{
	static uint8_t	col=0;
	uint8_t	shift;

	shift = 0x01 << col;	// Bitmaske erzeugen

	PORTC &= 0xE0;			// Abschalten aktuelle Reihe
	set_col(col);			// Setzen der neuen Reihe
	PORTC |= shift;			// Einschalten der neuen Reihe
	
	col++;					// Nchste Reihe
	if(col >= 5)
		col = 0;
} 


/*----------------------------------------------------------------------*/
//! Zurcksetzen der Ziffern-Bitfelder
//! \brief	Diese Funktion setzt die der Zeit entsprechende Anzahl 
/*----------------------------------------------------------------------*/
void zero_bitfield()
{

	sec_o.reg = 0;		// Sekunden Einstelle
	sec_o.bit_cnt = 0;
	
	sec_t.reg = 0;		// Sekunden Einstelle
	sec_t.bit_cnt = 0;

	minute_o.reg = 0;		// Sekunden Einstelle
	minute_o.bit_cnt = 0;

	minute_t.reg = 0;		// Sekunden Einstelle
	minute_t.bit_cnt = 0;

	hour_o.reg = 0;		// Sekunden Einstelle
	hour_o.bit_cnt = 0;

	hour_t.reg = 0;		// Sekunden Einstelle
	hour_t.bit_cnt = 0;									
}